import re
from typing import List

from functools import wraps
from collections import OrderedDict, defaultdict
from collections.abc import Sequence

from .serialization import get_json_serializer, type_of, TYPE_KEY

from .base import BaseDataClass

from .filtering import Filter, OrderBy

from spintop.utils.dict_ops import (
    update,
    deepen_dict,
    flatten_dict,
    tuplify_key
)

def record_call():
    def _record_call(fn):
        @wraps(fn)
        def _recorded(self, field_name, *args):
            self._record_call(fn.__name__, field_name, args)
            return fn(self, field_name, *args)
        return _recorded
    return _record_call

class QueryData(BaseDataClass):
    _calls : OrderedDict = OrderedDict

class QueryResult(BaseDataClass):
    rows: List[dict] = list
    query_id: str
    filters: List[Filter] = list
    orderby: List[OrderBy] = list
    query_type: str = None
    total_rows_count: int = None
    rows_count: int = None
    page_token: str = None
    next_page_token: str = None

class BaseQuery():

    def __init__(self, parts=[], *other_parts):
        # The field key must equal exactly value, or re.search if a compiled regex
        self._value_equals = dict()

        # The field of type list named key must contain value
        self._list_contains = dict()

        # The field named key must equal one of the sub value in the list value
        self._value_equals_one_of = dict()

        self._calls = OrderedDict()


        if not isinstance(parts, Sequence):
            parts = [parts]

        parts = list(self.default_parts()) + parts

        for query_part in tuple(parts) + other_parts:
            self.add(query_part)

    @property
    def value_equals(self):
        return self._value_equals

    @property
    def list_contains(self):
        return self._list_contains

    @property
    def value_equals_one_of(self):
        return self._value_equals_one_of

    def default_parts(self):
        """To override"""
        return []

    def add(self, query_part):
        query_part.inject_in_query(self)

    def _record_call(self, fn_name, field_name, arg):
        fn_name_and_key = self.KEY_SEPARATOR.join((fn_name, field_name))
        self._calls[fn_name_and_key] = arg

    def add_call(self, fn_name_and_key, args):
        fn_name, key = fn_name_and_key.split(self.KEY_SEPARATOR)
        getattr(self, fn_name)(key, *args)

    # Raw calls

    @record_call()
    def set_value_equals(self, field_name, value):
        self._value_equals[field_name] = value

    @record_call()
    def set_value_match(self, field_name, value):
        self._value_equals[field_name] = re.compile(value)

    @record_call()
    def set_value_one_of(self, field_name, *values):
        self._value_equals_one_of[field_name] = values

    @record_call()
    def set_list_contains(self, list_field_name, value):
        self._list_contains[list_field_name] = value

    @record_call()
    def clear(self, field_name, value=None):
        for _dict in (self.value_equals, self.value_equals_one_of, self.list_contains):
            _dict.pop(field_name, None)

    def __eq__(self, other_q):
        same_len = len(self._calls) == len(other_q._calls)
        if not same_len:
            return False

        for key, value in self._calls.items():
            if key not in other_q._calls or other_q._calls[key] != value:
                return False
        
        return True

    def __repr__(self):
        return '{}(eq={}, oneof={}, contains={})'.format(
            self.__class__.__name__,
            repr(self._value_equals),
            repr(self._value_equals_one_of),
            repr(self._list_contains)
        )

    KEY_SEPARATOR = '::'
    def as_dict(self):
        serializer = get_json_serializer()

        list_of_calls = serializer.serialize(QueryData(_calls=self._calls))['_calls']
        # TODO keep order of calls

        return {key: value for key, value in list_of_calls}

    @classmethod
    def from_dict(cls, _flat_separated_calls):
        query = cls()
        for fn_name, args in _flat_separated_calls.items():
            query.add_call(fn_name, args)
        return query

    def copy(self):
        return self.__class__.from_dict(self.as_dict())

def multi_query_serialize(**queries):
    """ Serialized as key_subkey = subvalue for key key, query in queries.
    """

    result = {}

    for key, query in queries.items():
        if BaseQuery.KEY_SEPARATOR in key:
            raise ValueError('Query key {!r} cannot contain {!r}.'.format(key, Query.KEY_SEPARATOR))
        if query:
            for subkey, subvalue in query.as_dict().items():
                result[BaseQuery.KEY_SEPARATOR.join((key, subkey))] = subvalue
    
    return result

def multi_query_deserialize(_dict):
    queries = {}

    for key, value in _dict.items():
        query_name, fn_name = split_key(key)
        update(queries, {
            query_name: {fn_name: value}
        })
    
    return {query_name: BaseQuery.from_dict(query_dict) for query_name, query_dict in queries.items()}

def split_key(key):
    split_key = key.split(BaseQuery.KEY_SEPARATOR)
    return split_key[0], BaseQuery.KEY_SEPARATOR.join(split_key[1:])


from collections import defaultdict
from typing import Any
import re

from .base import BaseDataClass

FILTER_EXPR_REGEX = r'([\w.]+)\s*([!><=~]{1,2})([\S\s]*)$'
OP_MAP = {
    '=': 'eq',
    '==': 'eq',
    '!=': 'neq',
    '<': 'lt',
    '<=': 'lte',
    '=<': 'lte',
    '>': 'gt',
    '>=': 'gte',
    '=>': 'gte',
    '~=': 'search'
}

ORDER_BY_EXPR_REGEX = r'([\w.]+)\s*(\+|-)'

class FilterFactory(object):
    def __init__(self, field_name):
        self.name = field_name

    def _filter(self, op, value):
        return Filter(
            name=self.name,
            op=op,
            val=value
        )
    
    def __eq__(self, value):
        return self._filter('eq', value)
        
    def __ne__(self, value):
        return self._filter('neq', value)

    def __lt__(self, value):
        return self._filter('lt', value)
        
    def __le__(self, value):
        return self._filter('lte', value)

    def __gt__(self, value):
        return self._filter('gt', value)

    def __ge__(self, value):
        return self._filter('gte', value)

    def match(self, regex):
        return self._filter('match', value)

    def search(self, regex):
        return self._filter('search', value)

class Filter(BaseDataClass):
    """A simple filter. Similar to SQLAlchemy filtrering system."""
    name: str
    op: str
    val: Any
    field_type: str = None

    @classmethod
    def from_expr(cls, expr):
        match = re.search(FILTER_EXPR_REGEX, expr)
        if not match:
            raise ValueError(f'Unable to parse expression {expr!r}')
        
        return cls(
            name=match.group(1).strip(),
            op=OP_MAP[match.group(2)],
            val=match.group(3).strip()
        )

    @classmethod
    def multi_from_declarative(cls, declarative_dict):
        """ With format:
        <name>:
            <op>: <val>
        <other name>:
            <op>: <val>
        ...
        """
        for name, op_val in declarative_dict.items():
            for op, val in op_val.items():
                yield cls(name=name, op=op, val=val)

    @staticmethod
    def multi_to_declarative(filters):
        result = defaultdict(dict)
        for f in filters:
            result[f.name][f.op] = f.val
        return result

    def inject_in_query(self, query):
        query.from_filters(self)

class OrderBy(BaseDataClass):
    name: str
    asc: bool
    
    @classmethod
    def from_expr(cls, expr):
        match = re.search(ORDER_BY_EXPR_REGEX, expr)
        if not match:
            raise ValueError(f'Unable to parse expression {expr!r}')
        
        asc = match.group(2).strip() == '+'

        return cls(
            name=match.group(1).strip(),
            asc=asc
        )


DEFAULT_OP_MAP_BY_FIELD_TYPE = {
    'string': ['eq', 'search'],
    'datetime': ['lt', 'gt']
}

class FilterSpec(BaseDataClass):
    name: str
    field_type: str
    input_type: str
    ops: list = list
    values: list = None
    label: str = None
    description: str = None

    def __post_init__(self):
        if not self.ops:
            self.ops = DEFAULT_OP_MAP_BY_FIELD_TYPE.get(self.field_type, ['eq'])


from .common import (
    NO_VERSION,
    VersionIDRecord,
    OutcomeData
)

from .base import (
    DefaultPrimitiveView,
    BaseDataClass,
    model_property,
    model_attribute,
    model_constant
)

from .filtering import (
    Filter,
    FilterFactory,
    FilterSpec,
    OrderBy
)

from .meta import FieldsOf

from .migration import DeprecatedBy

from .persistence import (
    PersistenceIDRecord,
    PersistenceRecord,
    PersistenceRecordCollection,
    SerializedPersistenceRecord,
    Query
)

from .queries import QueryResult, multi_query_deserialize, multi_query_serialize

from .serialization import (
    get_serializer, 
    get_bson_serializer, 
    get_json_serializer,
    type_of,
    serialized_type_of,
    is_type_of,
    type_dict_of,
    is_serialized_type_of,
    cls_of_serialized,
    ValidationError,
    AbstractSchema,
    base_type_mapping,
    register_schema_type,
    TYPE_KEY
)

from .view import DataClassPrimitiveView, ComplexPrimitiveView, update

from .view_helpers import (
    value_of_feature_named
)

from .test_records import *
from .lifecycles import *



import json

from typing import Callable, Type
from copy import copy
from contextlib import contextmanager
from dataclasses import dataclass

from spintop.analytics.base import REQUEST_ID, AbstractSingerTarget

from spintop.logs import _logger

ORG_ID_KEY = '_group'

logger = _logger('analytics')

@dataclass()
class ReservedStream():
    callback: Callable = None
    stream_type: Type = None
    stream_name: str = None

class ContextualAnalyticsTarget(AbstractSingerTarget):
    def __init__(self, analytics, reserved_streams=None, unlocked_streams=None, request_id=None, org_id=None):
        self.analytics = analytics
        if reserved_streams is None:
            reserved_streams = {}

        super().__init__()
        self.request_id = request_id
        self.org_id = org_id

        self.analytics = analytics
        self.reserved_streams = reserved_streams
        # self.unlocked_streams = unlocked_streams

    @property
    def reserved_stream_names(self):
        return set(self.reserved_streams.keys())

    def reserve_stream_name(self, name, callback=None, stream_type=None, stream_name=None):
        if stream_name is None:
            stream_name = name
        self.reserved_streams[name] = ReservedStream(callback=callback, stream_type=stream_type, stream_name=stream_name)

    def get_reserved_stream(self, name):
        return self.reserved_streams.get(name, ReservedStream())

    def unlock_target(self, unlock_stream_names=None, request_id=None, org_id=None):

        return self.__class__(
            self.analytics,
            reserved_streams=self.reserved_streams,
            unlocked_streams=unlock_stream_names,
            request_id=request_id,
            org_id=org_id
        )

    def update_named_stream(self, name, raw_data, stream_type=None):
        reserved_stream = self.get_reserved_stream(name)
        reserved_stream_type = reserved_stream.stream_type
        if stream_type and reserved_stream_type and reserved_stream_type != stream_type:
            raise ValueError(f'Mismatch between request stream type {stream_type!r} and reserved stream type {reserved_stream_type!r}')
        response = self.unlock_target({name})._update_named_endpoint_unlocked(reserved_stream.stream_name, raw_data, reserved_stream_type)
        if reserved_stream.callback:
            reserved_stream.callback(self.analytics.functions)
        return response

    def _update_named_endpoint_unlocked(self, name, raw_data, stream_type):
        return super().update_named_stream(name, raw_data, stream_type)

    @property
    def functions(self):
        return self.analytics.functions

    def send_messages_dict(self, messages_dict, request_id=None, stream_name=None):
        valid_messages, stream_names = self._validate_messages(messages_dict)
        response = self.analytics.send_messages_dict(
            valid_messages, 
            request_id=request_id if request_id else self.request_id, 
            stream_name=stream_name
        )
        return response

    def _validate_messages(self, messages_dict):
        validated_messages = []
        stream_names = set()
        for content in messages_dict:
            stream_name = content.get('stream')

            # self.assert_not_reserved(stream_name)
            
            stream_names.add(stream_name)
            content = self._message_transform(content)

            validated_messages.append(content)
        
        return validated_messages, stream_names
    
    def assert_not_reserved(self, stream_name):
        for reserved_stream in self.reserved_stream_names:
            if stream_name.startswith(reserved_stream):
                raise ValueError(f'Stream with name {stream_name!r} is reserved for internal use. No stream may start with {reserved_stream!r}.')

    def _message_transform(self, message_content):
        msg_type = message_content['type']
        if msg_type == 'SCHEMA':
            message_content['schema']['properties'][ORG_ID_KEY] = {'type': 'string'}
            message_content['key_properties'] = message_content.get('key_properties', []) + [ORG_ID_KEY]
        elif msg_type == 'RECORD':
            message_content['record'][ORG_ID_KEY] = self.org_id
            message_content['record'][REQUEST_ID] = self.request_id
        else:
            raise ValueError(f'Unsupported message type: {msg_type}')
        return message_content



import os
import sys
from glob import glob

from setuptools import setup, find_packages

HERE = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(HERE, 'spintop', 'VERSION')) as version_file:
    VERSION = version_file.read().strip()
packages = find_packages()

setup(
    name='spintop',
    version=VERSION,
    description='The python client to spintop.io',
    author='William Laroche',
    author_email='william.laroche@tackv.ca',
    maintainer='William Laroche',
    maintainer_email='william.laroche@tackv.ca',
    packages=packages,
    package_data={
        'spintop': ['VERSION']
    },
    install_requires=[
        'anytree',
        'appdirs',
        'click',
        'cryptography',
        'dataclasses>=0.6',
        'gitpython',
        'jinja2',
        'marshmallow>=3.5',
        'marshmallow-dataclass',
        'marshmallow-union',
        'marshmallow-jsonschema',
        'pkginfo',
        'PyJWT',
        'psutil',
        'pytz',
        'pyyaml',
        'python-dateutil',
        'requests',
        'tabulate',
        'tblib',
        'incremental-module-loader',
        'simplejson',
        'simple-memory-cache',
        'xmltodict'
    ],
    extras_require={
        'postgres': [
            'sqlalchemy',
            'psycopg2'
        ],
        'bigquery': [
            'jsonschema>=2.6.0',
            'singer-python>=1.5.0',
            'google-api-python-client>=1.6.2',
            'google-cloud>=0.34.0',
            'google-cloud-bigquery>=1.9.0',
            'oauth2client'
        ],
        'deploy': [
            'flask',
            'prefect',
            'google-cloud-secret-manager',
            'google-cloud-storage',
            'google-cloud-pubsub'
        ]
    },
    setup_requires=[
        'wheel>=0.29.0,<1.0',
    ],
    tests_require=[
        'mock>=2.0.0',
        'pytest>=2.9.2',
        'pytest-cov>=2.2.1',
    ],
)


from random import random

from prefect import task, Flow, case, Parameter
from prefect.tasks.control_flow import merge

from spintop_deploy.pub_sub_router.handlers import BigQueryJobHandler, GCSFileChangeHandler
from spintop_deploy.pub_sub_router.messages import BigQueryAuditLogMessage, GCSFileReferenceMessage


@BigQueryJobHandler.task()
def on_bq_message(message: BigQueryAuditLogMessage):
    return "This is a BQ message."

@GCSFileChangeHandler.task()
def on_gcs_file(message: GCSFileReferenceMessage):
    return "This is a gcs file."

with Flow("handle-message") as flow:
    message = Parameter('message')
    
    on_bq_message(message)
    on_gcs_file(message)


from typing import Optional

from spintop.models import (
    MeasureFeatureRecord,
    PhaseFeatureRecord,
)


class LCFeature(MeasureFeatureRecord, PhaseFeatureRecord):
    uuid: str 
    record_uuid: str
    pipeline_uuid: str = None
    # name: str

    # version: int = 0
    # index: int = 0
    # depth: int = 0
    # outcome: OutcomeData = None

    # value_f: Optional[float]
    # value_s: Optional[str]

    # user_data: dict = dict
    # feature_count: int = 0 # Number of children

    # def __post_init__(self):
    #     super().__post_init__()
    #     if self.outcome:
    #         self.outcome = OutcomeData.create(self.outcome)


import click
import subprocess

@click.group()
def cli():
    pass

@cli.group()
def prefect():
    pass

@prefect.command('exec')
def exec_prefect_app():
    subprocess.Popen("exec gunicorn --bind :$PORT --workers 1 --threads 1 --timeout 0 spintop_deploy.prefect_sls:app", shell=True).wait()


import os
import prefect
import time
import platform
from pprint import pprint


from flask import Flask, request

from .agent import ServerlessAgent

app = Flask(__name__)


flow_id = os.getenv('PREFECT__FLOW_ID')

@app.route('/', methods=['GET', 'POST'])
def hello_world():
    try:
        data = request.json
        print('POST content:')
        print(data)
    except Exception as e:
        print('Unable to get JSON: ' + str(e))

    ephemeral_agent = ServerlessAgent(name=f'sls-{platform.node()}', labels=[flow_id], show_flow_logs=True)
    ephemeral_agent.run_flow_id(flow_id, labels=[flow_id])
    print('DONE')
    return 'OK', 200

if __name__ == "__main__":
    app.run(debug=True, host='0.0.0.0', port=int(os.environ.get('PORT', 8080)))


import os

from tempfile import gettempdir
from shutil import copyfile

from spintop.models import BaseDataClass

from google.cloud import storage

class FileReference(BaseDataClass):
    name: str

    def download_to_filename(self, filename):
        copyfile(self.name, filename)
    
    def download_temp(self):
        filename = os.path.join(gettempdir(), self.bucket, self.name)

        folder, _ = os.path.split(filename)
        os.makedirs(folder, exist_ok=True)

        self.download_to_filename(filename)
        return filename

class GCSFileReference(FileReference):
    bucket: str

    def download_to_filename(self, filename):
        storage_client = storage.Client()

        bucket = storage_client.bucket(self.bucket)
        blob = bucket.blob(self.name)
        blob.download_to_filename(filename)
    
    def download_temp(self):
        filename = os.path.join(gettempdir(), self.bucket, self.name)

        folder, _ = os.path.split(filename)
        os.makedirs(folder, exist_ok=True)

        self.download_to_filename(filename)
        return filename

def list_bucket_files(bucket_name):
    storage_client = storage.Client()
    blobs = storage_client.list_blobs(bucket_name)
    return [GCSFileReference(bucket=bucket_name, name=blob.name) for blob in blobs]


import re

from typing import Dict, Any
from collections.abc import Mapping, Sequence, Mapping

from dataclasses import dataclass, fields, field, Field

from spintop.utils import GetRecursiveMixin, repr_obj

from .serialization import cls_of_serialized

def query_is(field_name, value):
    return QueryPart('set_value_equals', field_name, value)

def query_match(field_name, value):
    return QueryPart('set_value_match', field_name, value)

def query_is_one_of(field_name, values):
    return QueryPart('set_value_one_of', field_name, *values)

def query_list_contains(field_name, value):
    return QueryPart('set_list_contains', field_name, value)

def query_clear(field_name):
    return QueryPart('clear', field_name)

class QueryPart(object):
    def __init__(self, op, *args):
        self.op = op
        self.args = args

    def inject_in_query(self, query):
        fn = getattr(query, self.op)
        fn(*self.args)

class MultiQueryParts(object):
    def __init__(self, parts):
        self.parts = parts
    
    def inject_in_query(self, query):
        for part in self.parts:
            part.inject_in_query(query)

class FieldsOf(GetRecursiveMixin):
    def __init__(self, cls):
        self.type_ = cls
    
    def __iter__(self):
        for field in fields(self.type_):
            yield MetaField(self.type_, field)

    def __getattr__(self, key):
        field = self.get_field(key)
        return MetaField(self.type_, field)

    def get_field(self, key):
        if key == 'type_':
            raise RecursionError(key)
        cls = self.type_
        attr_value = getattr(cls, key, None)
        field = None
        
        try:
            fields_list = fields(cls)
        except TypeError:
            pass
        else:
            for possible_field in fields_list:
                if possible_field.name == key:
                    field = possible_field
                    break

        if isinstance(attr_value, property):
            try:
                field = PropertyField(attr_value, field=field)
            except TypeError as e:
                # add info
                raise TypeError(f'Property {key!r} of {cls} is invalid: {str(e)}')

        if field is None:
            raise AttributeError(f'No field named {key!r} in {cls}')
        else:
            return field

    def __eq__(self, other):
        try:
            return self.type_ == other.type_
        except AttributeError:
            return False

    def __repr__(self):
        return f'FieldsOf({self.type_.__name__})'

    def from_name(self, name):
        return self.get_recursive(name)

class CustomField():
    name = None
    type = None

    def accessor(self, name, obj):
        return getattr(obj, name)

    def setter(self, obj, value):
        return 

    def __repr__(self):
        return repr_obj(self, ['name', 'type'])

class PropertyField(CustomField):
    def __init__(self, property_attr, field=None):
        if field:
            self.name = field.name
            self.type = field.type
        else:
            self.name = property_attr.fget.__name__
            annotations = property_attr.fget.__annotations__

            if 'return' not in annotations:
                raise TypeError('Property field must declare the return type to be used as a meta field.')

            self.type = annotations['return']

class DictField(CustomField):
    def __init__(self, key):
        self.name = key
        self.type = Any

    def accessor(self, name, obj):
        return obj[name]

    def setter(self, obj, name, value):
        obj[name] = value

class MetaField():
    def __init__(self, cls, field, parents=()):
        self.__type = cls
        self.__parents = parents
        self.__field = field
    
    @property
    def field_name_(self):
        return self.__field.name

    @property
    def name_(self):
        return '.'.join(self.name_tuple_)

    @property
    def type_(self):
        return self.__field.type

    @property
    def field_(self):
        return self.__field

    @property
    def container_type_(self):
        return self.__type

    @property
    def root_container_type_(self):
        if self.__parents:
            return self.__parents[0].__type
        else:
            return self.__type

    @property
    def meta_fields_(self):
        return self.__parents + (self,)

    @property
    def name_tuple_(self):
        return tuple(meta_field.__field.name for meta_field in self.meta_fields_) 
    
    def __call__(self, obj):
        return self.get_value(obj)

    def get_value(self, obj):
        return self._get_value(
            obj, 
            context=dict(
                is_serialized=isinstance(obj, Mapping)
            )
        )

    def _get_value(self, obj, context, meta_fields=None):
        if meta_fields is None:
            meta_fields = self.meta_fields_

        for meta_field in meta_fields:
            accessor = meta_field.obj_accessor_factory(**context)
            obj = accessor(obj)
        return obj

    def set_value(self, obj, value):
        return self._set_value(
            obj, 
            value, 
            context=dict(
                is_serialized=isinstance(obj, Mapping)
            )
        )

    def _set_value(self, obj, value, context):
        meta_fields_except_last = self.meta_fields_[:-1]
        last_meta_field = self.meta_fields_[-1]
        if meta_fields_except_last:
            obj = self._get_value(obj, context, meta_fields=meta_fields_except_last)

        setter = last_meta_field.obj_setter_factory(**context)
        setter(obj, value)

    def obj_accessor_factory(self, is_serialized):
        if is_serialized:
            return self.serialized_accessor
        else:
            return self.obj_accessor

    def obj_accessor(self, obj): 
        self._check_type(obj.__class__)
        return getattr(obj, self.field_name_)

    def serialized_accessor(self, obj): 
        # No type check when serialized.
        # self._check_type(cls_of_serialized(obj))
        return obj[self.field_name_]   

    def _check_type(self, obj_cls):
        if obj_cls is None or not self.is_type_of_cls(obj_cls):
            raise AttributeError(f'Obj cls {obj_cls} does not match expected {self.container_type_}')

    def obj_setter_factory(self, is_serialized):
        if is_serialized:
            return self.serialized_setter
        else:
            return self.obj_setter
        
    def obj_setter(self, obj, value): 
        self._check_type(obj.__class__)
        setattr(obj, self.field_name_, value)

    def serialized_setter(self, obj, value):
        # No type check when serialized.
        obj[self.field_name_] = value

    def is_type_of_cls(self, cls):
        return cls is self.__type or issubclass(cls, self.__type)

    def is_container_of_cls(self, cls):
        return cls is self.root_container_type_ or issubclass(cls, self.root_container_type_)

    def __getattr__(self, key):
        cls = self.__field.type
        field = FieldsOf(cls).get_field(key)
        return self.__class__(cls, field, parents=self.meta_fields_)

    def __getitem__(self, key):
        cls = self.__field.type
        try:
            is_dict = issubclass(cls, Dict)
        except TypeError:
            is_dict = 'dict' in cls._name.lower()

        if is_dict:
            return DictMetaField(cls, DictField(key), parents=self.meta_fields_)
        else:
            try:
                return getattr(self, key)
            except AttributeError:
                raise KeyError(key)

    def __hash__(self):
        return hash(self.__field)

    def __repr__(self):
        return f'MetaField({self.root_container_type_.__name__}.{self.name_})'

    ## Query Building
    
    def __eq__(self, other):
        """When compared to another MetaField, check for equality and returns a Boolean.
        
        When compared to other types, returns a QueryPart to be used in a query."""
        if isinstance(other, MetaField):
            # Normal comparaison
            return self.__field == other.__field
        else:
            # Query comparaison
            query_value = other
            return self.is_(query_value)

    def __ge__(self, query_value):
        """For mappings: field >= {'x': 'y'}. 
        
        Contains at least the mapping exactly."""
        parts = self._build_mapping_parts(query_value)
        return MultiQueryParts(parts)

    def _build_mapping_parts(self, value):
        parts = []
        if isinstance(value, Mapping):
            for key, value in value.items():
                field = self[key]
                parts += field._build_mapping_parts(value)
        else:
            parts.append(self.__eq__(value))
        return parts

    def __le__(self, query_value):
        return self.one_of_(query_value)

    def is_(self, value):
        # value = getattr(value, , value)
        if hasattr(value, 'pattern'):
            # for regexes.
            return query_match(self.name_, value.pattern) # deconstruct re.compile to allow serialization
        else:
            return query_is(self.name_, value)

    def one_of_(self, value):
        return query_is_one_of(self.name_, value)

    def contains_(self, value):
        return query_list_contains(self.name_, value)

    def any_(self):
        return query_clear(self.name_)
    

class DictMetaField(MetaField):
    
    def obj_accessor(self, obj): 
        return obj[self.field_name_]

    def serialized_accessor(self, obj): 
        obj_cls = obj.__class__
        if not isinstance(obj, Mapping):
            raise AttributeError(f'Obj cls {obj_cls} should be a Mapping.')
        return obj[self.field_name_]   
        
    def obj_setter(self, obj, value): 
        obj[self.field_name_] = value

def _get_value_fields(obj, accessor, meta_fields):
    for meta_field in meta_fields:
        accessor_fn = getattr(meta_field, accessor)
        obj = accessor_fn(obj)
    return obj



import base64
import os


from contextlib import contextmanager

from spintop_deploy.logs import _logger

from .storage import GCSFileReference
from .pub_sub import get_encoder

logger = _logger('worker-handler')

@contextmanager
def cloud_functions_context(event, context):
    try:
        logger.info(
            'This Function was triggered by '
            f'message ID {context.event_id} of type {context.event_type} '
            f'published at {context.timestamp}'
        )
        os.environ['SPINTOP_DA_UUID'] = context.event_id
        yield
    except: # pylint: disable=bare-except
        logger.exception('Exception occured', exc_info=True)
        try_report_active_exc()

def try_report_active_exc():
    try:
        from google.cloud import error_reporting
        client = error_reporting.Client()
        client.report_exception()
    except: # pylint: disable=bare-except
        logger.warning('Unable to report to google error reporting', exc_info=True)

def pub_sub_handler_factory(callback):
    def process_message(event, context):
        """Background Cloud Function to be triggered by Pub/Sub.
        Args:
            event (dict):  The dictionary with data specific to this type of
            event. The `data` field contains the PubsubMessage message. The
            `attributes` field will contain custom attributes if there are any.
            context (google.cloud.functions.Context): The Cloud Functions event
            metadata. The `event_id` field contains the Pub/Sub message ID. The
            `timestamp` field contains the publish time.
        """
        with cloud_functions_context(event, context):
            raw_bytes = base64.b64decode(event['data'])
            message = get_encoder().decode(raw_bytes)
            callback(message)
    
    return process_message

def gcs_finalize_handler_factory(callback, download_file=False):

    def process_file(event, context):
        """Background Cloud Function to be triggered by Cloud Storage.
        This generic function logs relevant data when a file is changed.

        Args:
            event (dict): The Cloud Functions event payload.
            context (google.cloud.functions.Context): Metadata of triggering event.
        Returns:
            None; the output is written to Stackdriver Logging
        """
        
        # print('Event ID: {}'.format(context.event_id))
        # print('Event type: {}'.format(context.event_type))
        # print('Bucket: {}'.format(data['bucket']))
        # print('File: {}'.format(data['name']))
        # print('Metageneration: {}'.format(data['metageneration']))
        # print('Created: {}'.format(data['timeCreated']))
        # print('Updated: {}'.format(data['updated']))


        with cloud_functions_context(event, context):
            file_ref = GCSFileReference(name=event['name'], bucket=event['bucket'])
            if download_file:
                file_ref = file_ref.download_temp()
            callback(file_ref)
    
    return process_file

class HandlerRouter():
    def __init__(self):
        self.on_gcs_finalize = None
        self.on_gcs_delete = None
        self.on_pub_sub_publish = None

    def handler_factory(self):
        ROUTES = {
            'google.pubsub.topic.publish': self._handle_pub_sub_publish,
            'google.storage.object.finalize': self._handle_gcs_finalize
        }

        def _process(event, context):
            with cloud_functions_context(event, context):
                route = ROUTES.get(context.event_type, None)
                if not route:
                    raise NotImplementedError(f'HandlerRouter does not support {context.event_type}')
                route(event, context)

        return _process

    def register_on_gcs_finalize(self, fn):
        self.on_gcs_finalize = fn
        return fn

    def register_on_gcs_delete(self, fn):
        self.on_gcs_delete = fn
        return fn

    def register_on_pub_sub_publish(self, fn):
        self.on_pub_sub_publish = fn
        return fn

    def _handle_gcs_finalize(self, event, context):
        return self._handle_file(
            filename=event['name'], 
            bucket_name=event['bucket'], 
            handler_name='on_gcs_finalize'
        )

    def _handle_file(self, filename, bucket_name, handler_name):
        handler = getattr(self, handler_name, None)
        if not handler:
            logger.info(f'Handler {handler_name} not implemented.')
        else:
            file_ref = GCSFileReference(name=filename, bucket=bucket_name)
            return handler(file_ref)
    
    def _handle_pub_sub_publish(self, event, context):
        raw_bytes = base64.b64decode(event['data'])
        attributes = event.get('attributes', {})

        message = get_encoder().decode(raw_bytes)
        message_kind = message.get('kind', None)

        gcs_event_type = attributes.get('eventType', None)
        if message_kind and message_kind == 'storage#object':
            if gcs_event_type == 'OBJECT_FINALIZE':
                return self._handle_file(
                    filename=message['name'], 
                    bucket_name=message['bucket'], 
                    handler_name='on_gcs_finalize'
                )
            elif gcs_event_type == 'OBJECT_DELETE':
                return self._handle_file(
                    filename=message['name'], 
                    bucket_name=message['bucket'], 
                    handler_name='on_gcs_delete'
                )
            else:
                logger.info(f'GCS notification of type {gcs_event_type!r} not handled')
        else:
            handler = self.on_pub_sub_publish
            if not handler:
                raise NotImplementedError('Handler @register_on_pub_sub_publish not implemented.')
            return handler(message)


'''
Module for Singer literals and helpers.
'''
_PREFIX = '_sdc_'
RECEIVED_AT =      _PREFIX + 'received_at'
BATCHED_AT =       _PREFIX + 'batched_at'
SEQUENCE =         _PREFIX + 'sequence'
TABLE_VERSION =    _PREFIX + 'table_version'
PK =               _PREFIX + 'primary_key'
SOURCE_PK_PREFIX = _PREFIX + 'source_key_'
LEVEL_FMT =        _PREFIX + 'level_{}_id'
VALUE =            _PREFIX + 'value'


def parse_message_dict(obj):
    """Parse a message string into a Message object."""
    import singer
    import ciso8601

    from singer.messages import _required_key
    
    # We are not using Decimals for parsing here.
    # We recognize that exposes data to potentially
    # lossy conversions.  However, this will affect
    # very few data points and we have chosen to
    # leave conversion as is for now.
    msg_type = _required_key(obj, 'type')
    
    if msg_type == 'RECORD':
        time_extracted = obj.get('time_extracted')
        if time_extracted:
            try:
                time_extracted = ciso8601.parse_datetime(time_extracted)
            except:
                LOGGER.warning("unable to parse time_extracted with ciso8601 library")
                time_extracted = None


            # time_extracted = dateutil.parser.parse(time_extracted)
        return singer.RecordMessage(stream=_required_key(obj, 'stream'),
                             record=_required_key(obj, 'record'),
                             version=obj.get('version'),
                             time_extracted=time_extracted)


    elif msg_type == 'SCHEMA':
        return singer.SchemaMessage(stream=_required_key(obj, 'stream'),
                             schema=_required_key(obj, 'schema'),
                             key_properties=_required_key(obj, 'key_properties'),
                             bookmark_properties=obj.get('bookmark_properties'))

    elif msg_type == 'STATE':
        return singer.StateMessage(value=_required_key(obj, 'value'))

    elif msg_type == 'ACTIVATE_VERSION':
        return singer.ActivateVersionMessage(stream=_required_key(obj, 'stream'),
                                      version=_required_key(obj, 'version'))
    else:
        return None


import prefect
from prefect import case, task


from .messages import (
    GCS_FINALIZE, 
    GCS_DELETE, 
    GCSFileReferenceMessage, 
    BigQueryAuditLogMessage
)

def create_any_event(data=None, attributes=None):
    return dict(
        attributes=attributes,
        data=data
    )

class Handler(object):
    def __init__(self, callback):
        self.callback = callback

    @classmethod
    def task(cls, **task_kwargs):
        def _decorator(fn):
            obj = cls(fn)
            return prefect.tasks.core.function.FunctionTask(fn=obj._run_prefect, **task_kwargs)
        return _decorator

    def _run_prefect(self, message):
        handler_message = self.filter(message)
        if not handler_message:
            raise prefect.engine.signals.SKIP(f'Message filtered by handler {self.__class__.__name__}')
        else:
            self.handle(handler_message)

    def __call__(self, message):
        handler_message = self.filter(message)
        if handler_message:
            handler.handle(handler_message)

    def filter(self, message):
        return message

    def handle(self, message):
        return self.callback(message)

    def __repr__(self):
        return f'{self.__class__.__name__}(callback={self.callback!r})'

class GCSFileChangeHandler(Handler):
    @staticmethod
    def create_event(bucket, name, event_type):
        accepted_event_types = [GCS_FINALIZE, GCS_DELETE]

        if event_type not in accepted_event_types:
            raise ValueError(f'event_type must be either one of {accepted_event_types!r}')

        return create_any_event(
            attributes=dict(
                objectId=name,
                bucketId=bucket,
                eventType=event_type
            )
        )

    def filter(self, message):
        # This returns None if incompatible.
        return GCSFileReferenceMessage.from_message(message)


class BigQueryJobHandler(Handler):

    def filter(self, message):
        # This returns None if incompatible.
        return BigQueryAuditLogMessage.from_message(message)



from .base import SpintopAPIClientModule, SpintopAPIError
from .auth_bootstrap import (
    SpintopAPISpecAuthBootstrap, 
    SpintopAccessTokenDecoder,
    create_backend_auth_bootstrap_factory
)
from .analytics import SpintopAPIAnalytics
from .persistence_facade import SpintopAPIPersistenceFacade